/*
 Frequ systh for hawk 102-12 radio.
 DCP 20190416
 Libs as loaded from Etherkit Si5351 examples
 Try to add USB/LSB switch
 It looks like it works :)
 Try to add band limits. :)
 next, make 1Hz res.
 Add band memory. 
 Now, A6 is the band sensing.... 
 This is working but a case statement may be quicker than the /84 used.
 20191201 try to fix USB/LSB. Use clk1 as carrier gen in place ot 1.65Mhz crystal.
 further mods 20200224
 20210722 add Si5351 cal line.
 20230201 removed some commented out unused stuff.
 20230202 add fixes from Jeff WR2E and shorten the long push timeout.
 20230227 Speed up I2C as Jeff did
 20230228 Tyr to also use the underline as decade indicator. Works :)
 */

#define Version "HawkVFO_DCP_11SC"

#include "si5351.h"
#include "Wire.h"
#include <LiquidCrystal_I2C.h>

// synthesiser default settings on start up.
unsigned long frequency   =  0000000;          // 4 bytes required for frequency
unsigned long frequency0  =  0000000;          // 4 bytes required for frequency
unsigned long frequency1  =  1810000;          // 4 bytes required for frequency
unsigned long frequency2  =  3630000;          // 4 bytes required for frequency  
unsigned long frequency3  =  7150000;          // 4 bytes required for frequency
unsigned long frequency4  = 10125000;          // 4 bytes required for frequency
unsigned long frequency5  = 14100000;          // 4 bytes required for frequency
unsigned long frequency6  = 18068000;          // 4 bytes required for frequency
unsigned long frequency7  = 21000000;          // 4 bytes required for frequency
unsigned long frequency8  = 24000000;          // 4 bytes required for frequency
unsigned long frequency9  = 28000000;          // 4 bytes required for frequency
unsigned long frequency10 = 28000000;          // 4 bytes required for frequency

unsigned long step_size = 1000;             // default tuning increment (Hz)
unsigned long offset = 1650000;             // IF offset for clockgen when driving Kestral/Hawk RFDS Radio
static signed char ham_band_index = 0;      // start on 0hz

unsigned long USB =  1647000;           // carrier mixer for USB
unsigned long LSB =  1650000;           // carrier mixer for USB

int32_t cal_factor = 19100; //  eg. 16360 for my board, found from Si5351 calibrator sketch.

long encoder;                               // encoder direction: CCW=-1; NC=0; CW=1
Si5351 clockgen;                            // initalize clockgen

// encoder definitions
#define ENCODER_A  A0
#define ENCODER_B  A1
#define SSB_SWITCH A2                       // HIGH = LSB, LOW = USB.
#define AM_SWITCH  A3                       // not used yet  
// A4 = SDA
// A5 = SCL
#define BAND_SENSE     A6                  //
#define AnalogIn       A7                  // not used yet

#define Input1   2                        // not used yet
#define Input2   3                        // not used yet
#define Output2  4                        // not used yet
#define ENCODER_BUTTON 5                  // selects tuning decade
#define BAND_SWITCH    6                  // When held down, encoder selects band
#define Output1  13                       // not used yet
                           
#define ENCODER_INPUT_PORT PINC
int sideband = 0;   
int last_sideband = 1;

/* HF Ham bands
Band        UK Allocation          USA Allocation    Australia Allocation
(metres)          Mhz                   Mhz                 Mhz
 160         1.810 -  2.000        1.800 -  2.000        1.800 -  1.875
  80         3.500 -  3.800        3.500 -  4.000        3.500 -  3.800
  40         7.000 -  7.200        7.000 -  7.300        7.000 -  7.300
  30        10.100 - 10.150       10.100 - 10.150       10.100 - 10.150
  20        14.100 - 14.350       14.100 - 14.350       14.100 - 14.350
  17        18.068 - 18.168       18.068 - 18.168       18.068 - 18.168
  15        21.000 - 21.450       21.000 - 21.450       21.000 - 21.450 
  12        24.890 - 24.990       24.890 - 24.990       24.890 - 24.990
  10        28.000 - 29.700       28.000 - 29.700       28.000 - 29.700 
 */
 
//LiquidCrystal_I2C  lcd(0x20,4,5,6,0,1,2,3); // <<<<<<Works wih he GY-IICLCD board. "bl and blpol" not used.
//LiquidCrystal_I2C  lcd(0x3f,2,1,0,4,5,6,7,3,POSITIVE); // 0x3f is the I2C bus address for PCF8574AT HM board
//LiquidCrystal_I2C  lcd(0x27,2,1,0,4,5,6,7,3,POSITIVE); // 0x27 is the I2C bus address for PCF8574T HM board
LiquidCrystal_I2C lcd(0x27,16,2);  // set the LCD address to 0x30 maybe??


void setup() { 

  lcd.init();                      // initialize the lcd 
  lcd.backlight();     
  lcd.print(Version);
  delay(2000);
  lcd.clear();
  
  bool i2c_found;


  // Start serial and initialize the Si5351
  Serial.begin(57600);
//  i2c_found = clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.set_correction(cal_factor, SI5351_PLL_INPUT_XO); // <<<<<<<this is found from Si5351 calibrator code.

                        // make the I2C run at 400Khz instead of the default 100Khz.
Wire.setClock(400000);  // clockgen.init restes the I2C clock to 100khz so this needs to be after that.

  if(!i2c_found)
  {
    lcd.print("Device not found on I2C bus!");
  }
  
  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);     // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);  
  pinMode(ENCODER_BUTTON, INPUT_PULLUP);
  pinMode(BAND_SWITCH, INPUT_PULLUP);
  pinMode(SSB_SWITCH, INPUT_PULLUP);    // Hi = LSB, LOW =  USB
  pinMode(AM_SWITCH, INPUT_PULLUP);  
  pinMode(Input1, INPUT_PULLUP);
  pinMode(Input2, INPUT_PULLUP);
  pinMode(Output1,OUTPUT);
  pinMode(Output2,OUTPUT);  
  digitalWrite(Output1,HIGH);     // output1 off
  digitalWrite(Output2,HIGH);     // output2 off
  
  display_freq_step();            // display frequency & step-size
  set_synth();
}

/******************************
 main loop
*******************************/
void loop() {
  check_band();
  check_step_size();
  
if (digitalRead(SSB_SWITCH)!=last_sideband)
    { set_synth();
      display_freq_step();                        // display frequency & step-size
    }
  // check the frequency 
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    frequency = frequency + (encoder * step_size);
    
   set_synth();
    display_freq_step();                          // display frequency & step-size
  } 
}

/******************************
 read encoder direction
*******************************/
char read_encoder() {
  // definitions
  #define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce
 
  /* the array values are determined by combining the previous and current A1,A0 
  Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise 
  (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is 
  assigned to the remaining positions to indicate no movement*/   
  static char encoder[] = {0,-1,1,0,1,0,0,-1,-1,0,0,1,0,1,-1,0}; // SWAPPED PHASES
  //   static char encoder[] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};
  static unsigned char encoder_index = 0;
  
  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
  Typically there are four states (counts) per click, but often more due to contact bounce.
  Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant 
  when there is no rotation; and increases with CW rotation. Bounce can be completely 
  eliminated if we look for the "detent" code and integrate as shown below. */  
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];     
  
  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {                    
      integrator++;
    } 
  } else {                                       // encoder not resting on indent
     if (integrator > 0) {
       integrator--;
     }
  }
  
  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return(value);
}
  
/******************************
 check frequency band
*******************************/
void check_band() {               // band now set from A6 reading
  int temp = 0;
  temp = (analogRead(BAND_SENSE)/84);
    
if (ham_band_index!=temp){    // update the band if it has changed
        ham_band_index=temp;
  unsigned long ham_band[] = {frequency0,frequency1,frequency2,frequency3,frequency4,frequency5,frequency6,frequency7,frequency8,frequency9,frequency10};
        frequency = ham_band[ham_band_index];     // change frequency band  
        set_synth ();                             // set the new frequ ****** added 20190502
        display_freq_step(); 
}
}


/******************************
 check the step size
*******************************/
void check_step_size() {
  // delay definitions
  #define DELAY 3000                             // DELAY = debounce(mS)*loops/second
  #define LONG_PUSH 5000                         // number of loops considered "long"
  
  // variables
  static boolean transition = false;             // valid during button push
  static unsigned short integrator;              // varies from 0 to MAXIMUM
  static unsigned short push_duration;           // determines step size/direction
  
  // look for first high-to-low transition
  if (!transition && (digitalRead(ENCODER_BUTTON) == LOW)) {
    transition = true;
    integrator = 0;
    push_duration = 0;
  }
  
  // look for last low-to-high transition
  if (transition) {
    if (push_duration < (LONG_PUSH*2)) {         // track push duration
      push_duration++;
    }
    if (digitalRead(ENCODER_BUTTON) == HIGH) {   // ENCODER_BUTTON is HIGH
      if (integrator < DELAY) {        
        integrator++;
      } else {                                   // ENCODER_BUTTON is LOW
        if (integrator > 0) {
    integrator--;
  }
      }
    }
   
    // update the step_size
    if (integrator >= DELAY) { 
      transition = false;
      if (push_duration > LONG_PUSH) {          // LONG PUSH: decrease step size
        step_size = step_size/10;
        if (step_size < 1) {                    // was 10
          step_size = 100000;                   // wrap around    
        }        
      } else {                                  // SHORT PUSH: increase step size
        step_size = step_size*10;
        if (step_size > 100000) {
          step_size = 1;                       // wrap around Was 10
        }         
      }
      
      display_freq_step();                      // update step-size     
    }
  }
}


/*******************************
set the Si5351 CLK0 frequency
 ******************************/
void set_synth ()
{
switch (ham_band_index) {       // limit to SSB part of bands

    case 0:
  frequency = constrain(frequency, 00000000, 00000000);
  frequency0 = frequency;
  break;

    case 1:
  frequency = constrain(frequency, 1840000, 1875000);
  frequency1 = frequency;
  break;

      case 2:
  frequency = constrain(frequency, 3525000, 3700000);
  frequency2 = frequency;
  break; 

    case 3:
  frequency = constrain(frequency, 7040000, 7300000);
  frequency3 = frequency;
  break;

    case 4:
  frequency = constrain(frequency, 10115000, 10140000);
  frequency4 = frequency;
  break;

    case 5:
  frequency = constrain(frequency, 14112000, 14350000);
  frequency5 = frequency;
  break;

    case 6:
  frequency = constrain(frequency, 18110000, 18168000);
  frequency6 = frequency;
  break;

    case 7:
  frequency = constrain(frequency, 21150000, 21450000);
  frequency7 = frequency;
  break;

    case 8:
  frequency = constrain(frequency, 00000000, 00000000);
  frequency8 = frequency;
  break;
  
    case 9:
  frequency = constrain(frequency, 00000000, 00000000);
  frequency9 = frequency;
  break;
  
     case 10:
  frequency = constrain(frequency, 00000000, 00000000);
  frequency10 = frequency;
  break;
 
}

    if (digitalRead(SSB_SWITCH)==LOW) // USB
    {
      clockgen.set_freq((USB*100), SI5351_CLK1);
      offset=USB;
    last_sideband=0;
    }
  else                                // LSB
    {
      clockgen.set_freq((LSB*100), SI5351_CLK1);
      offset=LSB;
    last_sideband=1;
    }
    // set the frequency
    clockgen.set_freq(((frequency-offset)*100), SI5351_CLK0);
}

/********************************
 update frequency and step-size
********************************/
void display_freq_step() {
  long MHz = frequency/1000000;
   
  // convert frequency to floating point
  // adding 0.0000001 fixes a rounding error
  double freq = (double) frequency/1000000+0.0000001;

  // display frequency
  lcd.clear();
      if (digitalRead(SSB_SWITCH)==HIGH)
      { lcd.print("L");
      } 
      else 
      { lcd.print("U");
      }
  
  if (MHz < 10) {
    lcd.setCursor(3,0);                          // start at char 1, line 1
    lcd.print(freq, 6);                          // 6 digit resolution
    lcd.print(" MHz ");
  } else {
    lcd.setCursor(2,0);                          // start at char 1, line 1
    lcd.print(freq, 6);
    lcd.print(" MHz ");
  }

  // display step size
  switch (step_size) {
    
      case 1:                           // added for 1xHz.
      lcd.setCursor(10,0);
      lcd.cursor();       
      break;    
    case 10:
      lcd.setCursor(9,0);
      lcd.cursor();     
      break;
    case 100:
      lcd.setCursor(8,0);
      lcd.cursor();        
      break;
    case 1000:
      lcd.setCursor(7,0);
      lcd.cursor();         
      break;
    case 10000:
      lcd.setCursor(6,0);
      lcd.cursor();        
      break;
    case 100000:
      lcd.setCursor(5,0);
      lcd.cursor();     
      break;
    default:
      break;
   lcd.noCursor();     

  }
}
